
#include "Arduino.h"
// bras de pont phase A
#define INA_1 11 // transistor haut
#define INA_2 12 // transistor bas
// bras de pont phase B
#define INB_1 5 
#define INB_2 2 
// bras de pont phase C
#define INC_1 6 
#define INC_2 7 

#define PWM
#define TRIGGER 3 // PE5
#define CONTROL 4 // PG5

#define HIN_EN 14
#define HIN_IN1 15
#define HIN_IN2 16

// états d'un bras de pont
#define HIGH_0_LOW_1 0 // sortie à la masse (état 01)
#define HIGH_1_LOW_0 1 // sortie en PWM (état 10)
#define HIGH_0_LOW_0 2 // sortie non connectée (état 00)

uint8_t sequence_A[6] = {1,1,2,0,0,2};
uint8_t sequence_B[6] = {0,2,1,1,2,0};
uint8_t sequence_C[6] = {2,0,0,2,1,1};

uint8_t sensRot = 0;
uint8_t etat_capteurs_anterieur;

// capteurs HALL
#define IN1 18 // PD3
#define IN2 19 // PD2
#define IN3 20 // PD1
uint8_t tableEtats[8];
uint32_t compteur;
uint32_t temps;
float period;
float omega;
float omegaDemand = 1.0;
float Kp=0.02;
float Ki=0.02;
float integOmega;
#define BUFSIZE 10
float bufPeriod[BUFSIZE];
uint8_t ibuf;
uint8_t etatControl;
bool commut = true;
uint32_t updateTime;
bool rotation;

uint16_t icr;
uint16_t ocra,ocrb;
uint16_t temps_mort = 10;
uint16_t diviseur[6] = {0,1,8,64,256,1024};

float periodPWM = 50;
float rapport = 0.0;
uint8_t mode = 0; // mode de fonctionnement  : 0 : contrôle du rapport cyclique, 1 : contrôle de la vitesse
uint8_t alim = 0; // alimentation 1 : moteur, 2 : roue libre, 3 : freinage
#define RAPPORT_MIN 0.05
#define RAPPORT_MAX 0.7

// communication Série
#define GET_DATA 10
#define SET_DATA 11
#define DATA_0_SIZE 4 // rapport cyclique (float)
#define DATA_1_SIZE 1 // sens de rotation (int8)
#define DATA_2_SIZE 4 // période en us (float)
#define DATA_3_SIZE 1 // mode de fonctionnement (int8)
#define DATA_4_SIZE 4*BUFSIZE // tampon des périodes 
#define DATA_5_SIZE 4 // vitesse en tours/s (float)
#define DATA_6_SIZE 1 // alimentation 1 : moteur, 2 : roue libre, 3 : freinage
#define DATA_7_SIZE 4 // Kp
#define DATA_8_SIZE 4 // Ki
uint8_t data_0[DATA_0_SIZE];
uint8_t data_1[DATA_1_SIZE];
uint8_t data_2[DATA_2_SIZE];
uint8_t data_3[DATA_3_SIZE];
uint8_t data_5[DATA_5_SIZE];
uint8_t data_6[DATA_6_SIZE];
uint8_t data_7[DATA_7_SIZE];
uint8_t data_8[DATA_8_SIZE];

bool data_2_ready = true;
bool data_2_request = false;
bool data_4_ready = false;
bool data_4_request = false;

void config_timers(float period, float rapport) {
  // period : période du PWM en microsecondes
  cli();
  TCCR1A = (1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0);
  TCCR1B = (1 << WGM13);
  TCCR3A = (1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0);
  TCCR3B = (1 << WGM13);
  TCCR4A = (1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0);
  TCCR4B = (1 << WGM13);

  icr = (F_CPU/1000000*period/2);
    int d = 1;
    while ((icr>0xFFFF)&&(d<5)) {
        d++;
        icr = (F_CPU/1000000*period/2/diviseur[d]);
    } 
  TCCR1B |= d;
  TCCR3B |= d;
  TCCR4B |= d;
  if (icr*rapport>temps_mort) {
    ocra = icr*rapport-temps_mort;
    ocrb = icr*rapport+temps_mort;
  }
  else {
    ocra = icr*rapport;
    ocrb = icr*rapport;
  }
  TCNT1 = 0;
  TCNT3 = 0;
  TCNT4 = 0;
  ICR1 = icr;
  ICR3 = icr;
  ICR4 = icr;
  OCR1A = ocra;
  OCR1B = ocrb;
  OCR3A = ocra;
  OCR3B = ocrb;
  OCR4A = ocra;
  OCR4B = ocrb;
  sei();
}


void stop_pwm() {
  TCCR1A &= ~((1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0));
  TCCR3A &= ~((1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0));
  TCCR4A &= ~((1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0));
}

void start_pwm() {
  TCCR1A = (1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0);
  TCCR3A = (1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0);
  TCCR4A = (1 << COM1A1) | (1 << COM1B1) | (1 << COM1B0);
}

void open_all() {
  digitalWrite(INA_1,LOW);
  digitalWrite(INA_2,LOW);
  digitalWrite(INB_1,LOW);
  digitalWrite(INC_2,LOW);
  digitalWrite(INC_1,LOW);
  digitalWrite(INC_2,LOW);
}

void changer_rapport_pwm(float rapport) {
  if (icr*rapport>temps_mort) {
    ocra = icr*rapport-temps_mort;
    ocrb = icr*rapport+temps_mort;
  }
  else {
    ocra = icr*rapport;
    ocrb = icr*rapport;
  }
  
}

void etat_bras_A(uint8_t etat) {
#ifdef PWM
  switch (etat) {
    case HIGH_0_LOW_1:
      OCR1A = 0;
      OCR1B = 0;
      break;
    case HIGH_1_LOW_0:
      OCR1A = ocra;
      OCR1B = ocrb;
      break;
    case HIGH_0_LOW_0:
      OCR1A = 0;
      OCR1B = icr+1;
      break;
  }
  #else
    switch (etat) {
    case HIGH_0_LOW_1:
      digitalWrite(INA_1,LOW);
      digitalWrite(INA_2,HIGH);
      break;
    case HIGH_1_LOW_0:
      digitalWrite(INA_1,HIGH);
      digitalWrite(INA_2,LOW);
      break;
    case HIGH_0_LOW_0:
      digitalWrite(INA_1,LOW);
      digitalWrite(INA_2,LOW);
      break;
  }
  #endif
}

void etat_bras_B(uint8_t etat) {
#ifdef PWM
  switch (etat) {
    case HIGH_0_LOW_1:
      OCR3A = 0;
      OCR3B = 0;
      break;
    case HIGH_1_LOW_0:
      OCR3A = ocra;
      OCR3B = ocrb;
      break;
    case HIGH_0_LOW_0:
      OCR3A = 0;
      OCR3B = icr+1;
      break;
  }
#else
  switch (etat) {
    case HIGH_0_LOW_1:
      digitalWrite(INB_1,LOW);
      digitalWrite(INB_2,HIGH);
      break;
    case HIGH_1_LOW_0:
      digitalWrite(INB_1,HIGH);
      digitalWrite(INB_2,LOW);
      break;
    case HIGH_0_LOW_0:
      digitalWrite(INB_1,LOW);
      digitalWrite(INB_2,LOW);
      break;
  }
#endif
}

void etat_bras_C(uint8_t etat) {
#ifdef PWM
  switch (etat) {
    case HIGH_0_LOW_1:
      OCR4A = 0;
      OCR4B = 0;
      break;
    case HIGH_1_LOW_0:
      OCR4A = ocra;
      OCR4B = ocrb;
      break;
    case HIGH_0_LOW_0:
      OCR4A = 0;
      OCR4B = icr+1;
      break;
  }
#else
  switch (etat) {
    case HIGH_0_LOW_1:
      digitalWrite(INC_1,LOW);
      digitalWrite(INC_2,HIGH);
      break;
    case HIGH_1_LOW_0:
      digitalWrite(INC_1,HIGH);
      digitalWrite(INC_2,LOW);
      break;
    case HIGH_0_LOW_0:
      digitalWrite(INC_1,LOW);
      digitalWrite(INC_2,LOW);
      break;
  }
#endif
}

void change_rapport() { 
        if (omega!=0) integOmega += (omegaDemand-omega)/omega;
        else integOmega += omegaDemand*0.2;
        rapport = Kp*(omegaDemand-omega)+Ki*integOmega;
        if (rapport>RAPPORT_MAX) rapport = RAPPORT_MAX;
        if (rapport<0) rapport = RAPPORT_MIN;
        changer_rapport_pwm(rapport);
}

void change_etat() {
  if (commut==false) return;
  uint32_t t;
  uint8_t etat_capteurs = (PIND >> 1)&0b111;
  
  if (omega!=0) {
    if (etat_capteurs==etat_capteurs_anterieur) return;
    etat_capteurs_anterieur = etat_capteurs;
  }
  
  
  if ((etat_capteurs==0)||(etat_capteurs==0xb111)) return;
  uint8_t is = tableEtats[etat_capteurs];
  etat_bras_A(sequence_A[is]);
  etat_bras_B(sequence_B[is]);
  etat_bras_C(sequence_C[is]);

  compteur += 1;
  if (compteur>=2) rotation =true;
  if (compteur==24) { // 1 tour
    
    if (etatControl==0) {
      PORTG |= (1<<5); //digitalWrite(CONTROL,HIGH);
      etatControl = 1;
    }
    else {
      PORTG &= ~(1<<5); //digitalWrite(CONTROL,LOW);
      etatControl = 0;
    }
    compteur = 0;
    t = micros();
    period = t-temps;
    temps = t;
    bufPeriod[ibuf] = period;
    PORTE &= ~(1<<5); //digitalWrite(TRIGGER,LOW);
    if (mode==1) {
      omega = 1/(period*1e-6);
      
      change_rapport();
    }
    ibuf++;
    if (ibuf==BUFSIZE) {
      ibuf = 0;
      data_4_ready = true;
    }
    memcpy(data_2,&period,DATA_2_SIZE);
    data_2_ready = true;
    
  }

}

void remplir_table_etats(uint8_t sensRot) {
  if (sensRot==0) {
    tableEtats[B010] = 3;
    tableEtats[B011] = 4;
    tableEtats[B001] = 5;
    tableEtats[B101] = 0;
    tableEtats[B100] = 1;
    tableEtats[B110] = 2;
  }
  else {
    tableEtats[B010] = 0;
    tableEtats[B011] = 1;
    tableEtats[B001] = 2;
    tableEtats[B101] = 3;
    tableEtats[B100] = 4;
    tableEtats[B110] = 5;
  }

}

void get_data() {
  char n;
  while (Serial.available()<1) {};
  n = Serial.read();
  if (n==2) data_2_request = true;
  if (n==4) data_4_request = true;
}

void send_data() {
  if ((data_2_ready)&&(data_2_request)) {
      //data_2_ready = false;
      data_2_request = false;
      Serial.write(data_2,DATA_2_SIZE);
  }
  if ((data_4_ready)&&(data_4_request)) {
      //data_4_ready = false;
      data_4_request = false;
      Serial.write((uint8_t *)bufPeriod,DATA_4_SIZE);
  }
}	

void set_data() {
  char n;
  while (Serial.available()<1) {};
  n = Serial.read();
  if (n==0) {
    while (Serial.available()<DATA_0_SIZE) {};
    Serial.readBytes(data_0,DATA_0_SIZE);
    memcpy(&rapport,data_0,DATA_0_SIZE);
    //digitalWrite(TRIGGER,HIGH);
    
    changer_rapport_pwm(rapport);
    period = 0.0;
    change_etat();
  }
  else if (n==1) {
    while (Serial.available()<DATA_1_SIZE) {};
    Serial.readBytes(data_1,DATA_1_SIZE);
    memcpy(&sensRot,data_1,DATA_1_SIZE);
    remplir_table_etats(sensRot);
  }
  
  else if (n==3) {
    while (Serial.available()<DATA_3_SIZE) {};
    Serial.readBytes(data_3,DATA_3_SIZE);
    memcpy(&mode,data_3,DATA_3_SIZE);
    
  }
  else if (n==5) { // vitesse angulaire demandée
    while (Serial.available()<DATA_5_SIZE) {};
    Serial.readBytes(data_5,DATA_5_SIZE);
    memcpy(&omegaDemand,data_5,DATA_5_SIZE);
    digitalWrite(TRIGGER,HIGH);
    change_rapport();
    change_etat();
  }
  else if (n==6) { // pilotage du pont d'alimentation
    while (Serial.available()<DATA_6_SIZE) {};
    Serial.readBytes(data_6,DATA_6_SIZE);
    memcpy(&alim,data_6,DATA_6_SIZE);
    digitalWrite(HIN_EN,LOW);
    if (alim==1) { // alimentation connectée
      commut = true;
      start_pwm();
      digitalWrite(HIN_IN1,LOW);
      digitalWrite(HIN_IN2,HIGH);
    }
    else if (alim==2) { // alimentation déconnectée, roue libre
      digitalWrite(TRIGGER,LOW);
      digitalWrite(TRIGGER,HIGH);
      //commut = false; // à décommenter pour obtenir une roue libre sans commutation
      //stop_pwm();
      //open_all();
      
      digitalWrite(HIN_IN1,LOW);
      digitalWrite(HIN_IN2,LOW);
      //changer_rapport_pwm(0.7);

    }
    else if (alim==3) { // alimentation déconnectée, frein moteur
      commut = true;
      start_pwm();
      digitalWrite(HIN_IN1,HIGH);
      digitalWrite(HIN_IN2,LOW);
      
      digitalWrite(TRIGGER,LOW);
    digitalWrite(TRIGGER,HIGH);
    }
    digitalWrite(HIN_EN,HIGH);
  }
  else if (n==7) {
    while (Serial.available()<DATA_7_SIZE) {};
    Serial.readBytes(data_7,DATA_7_SIZE);
    memcpy(&Kp,data_7,DATA_7_SIZE);
  }
  else if (n==8) {
    while (Serial.available()<DATA_8_SIZE) {};
    Serial.readBytes(data_8,DATA_8_SIZE);
    memcpy(&Ki,data_8,DATA_8_SIZE);
  }
  
}

void read_serial() {
   char com;
   if (Serial.available()>0) {
        com = Serial.read();
        if (com==GET_DATA) get_data();
        else if (com==SET_DATA) set_data();
   }
}	


void setup() {
  Serial.begin(115200);
  while(!Serial);  
  remplir_table_etats(sensRot);
  pinMode(INA_1,OUTPUT);
  digitalWrite(INA_1,LOW);
  pinMode(INA_2,OUTPUT);
  digitalWrite(INA_2,LOW);
  pinMode(INB_1,OUTPUT);
  digitalWrite(INB_1,LOW);
  pinMode(INB_2,OUTPUT);
  digitalWrite(INB_2,LOW);
  pinMode(INC_1,OUTPUT);
  digitalWrite(INC_1,LOW);
  pinMode(INC_2,OUTPUT);
  digitalWrite(INC_2,LOW);

  pinMode(IN1,INPUT);
  pinMode(IN2,INPUT);
  pinMode(IN3,INPUT);

  pinMode(TRIGGER,OUTPUT);
  digitalWrite(TRIGGER,LOW);
  pinMode(CONTROL,OUTPUT);
  digitalWrite(CONTROL,LOW);

  pinMode(HIN_EN,OUTPUT);
  pinMode(HIN_IN1,OUTPUT);
  pinMode(HIN_IN2,OUTPUT);
  digitalWrite(HIN_IN1,LOW);
  digitalWrite(HIN_IN2,HIGH);
  digitalWrite(HIN_EN,HIGH);
 
  attachInterrupt(digitalPinToInterrupt(IN1),change_etat,CHANGE);
  attachInterrupt(digitalPinToInterrupt(IN2),change_etat,CHANGE);
  attachInterrupt(digitalPinToInterrupt(IN3),change_etat,CHANGE);
  updateTime = millis();
#ifdef PWM
  config_timers(periodPWM,rapport);
#endif
  ibuf = 0;
  compteur = 0;
  temps = micros();
  omega = 0;
  integOmega = 0;
  etatControl = 0;
  etat_capteurs_anterieur = 0;
  commut = true;
  rotation = false;
  change_etat();

}

void loop() {
  read_serial();
  send_data();
  delay(100);
  if ((mode==1)&&(omega==0)) {
    change_rapport();
    change_etat();
  }
}

                